@@ -73,7 +73,8 @@ const char * const longopts_help[] = {
 };
 const char *usagestring = NULL;
 
-static int get_dirfd(char *path, bool symlinks) {
+static int get_dirfd(char *path, bool symlinks)
+{
 	char *ch;
 	char *item;
 	char *linkpath = NULL;
@@ -150,7 +151,7 @@ static char *clean_path(char *path)
 	char *ch;
 	char *ch2;
 	char *str;
-	str = xmalloc(strlen(path));
+	str = xmalloc(strlen(path) + 1);
 	ch = path;
 	ch2 = str;
 	while (true) {
@@ -276,7 +277,7 @@ static int do_check(char *path, uid_t uid, gid_t gid, mode_t mode,
 
 		if (mode && (st.st_mode & 0777) != mode) {
 			if ((type != inode_dir) && (st.st_nlink > 1)) {
-				eerror("%s: chmod: %s %s", applet, "Too many hard links to", path);
+				eerror("%s: chmod: Too many hard links to %s", applet, path);
 				close(readfd);
 				return -1;
 			}
